/*
 * Copyright (C) 2006 Erik Johansson <erik@ejohansson.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _POSIX_C_SOURCE 199309
#include <time.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <fcntl.h>

#include <linux/input.h>

#include <syslog.h>

int main(int argc, char **argv)
{
	if (argc != 3) {
		printf("Usage: %s <keyboard> <multimedia>\n", argv[0]);
		printf("       <keyboard>   is e.g. /dev/input/event1\n");
		printf("       <multimedia> is e.g. /dev/input/event2\n\n");
		printf("This program reads volume knob events from <multimedia> and\n");
		printf("writes them as key press events to <keyboard>. Therefor, you'll\n");
		printf("need write access to <keyboard> and read access to <multimedia>.\n\n");

		printf("See /proc/bus/input/devices to find which devices (Handlers) to use.\n");
		printf("<keyboard> is the device for your keyboard's standard keys.\n");
		printf("<multimedia> is the device for your keyboard's multimedia keys.\n\n");

		printf("Running \"cat <keyboard>\" should give output when you press the space bar.\n");
		printf("Running \"cat <multimedia>\" should give output when turn the volume knob.\n");
		exit(EXIT_FAILURE);
	}

	openlog("knob", LOG_PID | LOG_PERROR, LOG_LOCAL0);

	int keyboard = open(argv[1], O_WRONLY);
	if (keyboard < 0) {
		syslog(LOG_ERR, "Unable to open keyboard device %s: %m", argv[1]);
		exit(EXIT_FAILURE);
	}

	struct pollfd multimedia = { .events = POLLIN };
	multimedia.fd = open(argv[2], O_RDONLY);
	if (multimedia.fd < 0) {
		syslog(LOG_ERR, "Unable to open multimedia device %s: %m", argv[2]);
		exit(EXIT_FAILURE);
	}

	// Fork off the parent process
	pid_t pid = fork();
	if (pid < 0) {
		syslog(LOG_ERR, "Failed to fork off the parent process. %m");
		exit(EXIT_FAILURE);
	}
	// If we got a good PID, then we can exit the parent process.
	else if (pid > 0) {
		exit(EXIT_SUCCESS);
	}

	// Change the file mode mask.
	umask(0);

	// Create a new SID for the child process.
	if (setsid() < 0) {
		syslog(LOG_ERR, "Failed to create a new SID for the child process. %m");
		exit(EXIT_FAILURE);
	}

	// Change the current working directory.
	if (chdir("/") < 0) {
		syslog(LOG_ERR, "Failed to change current working directory. %m");
		exit(EXIT_FAILURE);
	}

	// Don't print any subsequent errors to stderr
	closelog();
	openlog("knob", LOG_PID, LOG_LOCAL0);

	// Close stdin, stdout and stderr
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	syslog(LOG_INFO, "Keyboard device: %s, Multimedia device: %s", argv[1], argv[2]);

	struct input_event event;
	struct input_event send = { .type = EV_KEY };

	// Time between key down and key up
	const struct timespec keypress_time = { .tv_sec = 0, .tv_nsec = 1000 };

	while (1) {
		if (poll(&multimedia, 1, -1) < 0) {
			syslog(LOG_ERR, "Error in poll (%s)", argv[2]);
			exit(EXIT_FAILURE);
		}

		const size_t rb = read(multimedia.fd, &event, sizeof(struct input_event));
		if (rb != sizeof(struct input_event)) {
			syslog(LOG_ERR, "Error in read (%s)", argv[2]);
			exit(EXIT_FAILURE);
		}

		if (event.type == EV_ABS && event.value != 0) {
			send.code = (event.value < 0) ? KEY_VOLUMEDOWN : KEY_VOLUMEUP;
			send.time = event.time;

			for (int db = abs(event.value); db > 0; db--) {
				// Simulate keypress
				send.value = 1;
				size_t wb = write(keyboard, &send, sizeof(struct input_event));
				if (wb != sizeof(struct input_event)) {
					syslog(LOG_ERR, "Error in key down write (%s)", argv[1]);
					exit(EXIT_FAILURE);
				}

				nanosleep(&keypress_time, NULL);

				// Simulate keyrelease
				send.value = 0;
				wb = write(keyboard, &send, sizeof(struct input_event));
				if (wb != sizeof(struct input_event)) {
					syslog(LOG_ERR, "Error in key up write (%s)", argv[1]);
					exit(EXIT_FAILURE);
				}

				nanosleep(&keypress_time, NULL);
			}
		}
	}

	closelog();
	return EXIT_SUCCESS;
}
